/***************************************************************************
                          k8000.c  -  description
                             -------------------
    begin                : Mon May 29 2000
    email                : j.struyve@pandora.be
 ***************************************************************************/

/***************************************************************************
 *   This code is ported to Linux starting with the DOS routines delivered *
 *   by Velleman grouped into a RTLinux kernel module.                     *
 *                                                                         *
 ***************************************************************************/

/* CVS information :
 *
 * $Id: mod_k8000.c,v 1.3 2001/05/20 08:24:41 js Exp $
 *
 */

#define MODULE
#define __KERNEL__
#define MODVERSIONS

#include <linux/kernel.h>
#include <linux/module.h>
#include <asm/io.h>
#include "mod_k8000.h"

#define MaxIOcard 3
#define MaxIOchip 7
#define MaxIOchannel 64
#define MaxDACchannel 32
#define MaxADchannel 16
#define MaxDAchannel 4

extern char *Bin;

short IOchipCode[] = {0x70, 0x72, 0x74, 0x76, 0x78, 0x7A, 0x7C, 0x7E};
short DACchipCode[] = {0x40, 0x42, 0x44, 0x46};
short ADDAchipCode[] = {0x90, 0x92, 0x94, 0x96};

short IOconfig[MaxIOchip+1];
short IOdata[MaxIOchip+1];
short IO[MaxIOchannel+1];
short DAC[MaxDACchannel+1];
short AD[MaxADchannel+1];
short DA[MaxDAchannel+1];
short ControlPort;
short StatusPort;
short I2CbusDelay;

/*
	***********************
	* Kernel module part  *
	***********************
*/

int init_module(void)
{
	/* do nothing, just load the functions */
	return 0;
}

void cleanup_module(void)
{/* there's nothing to clean*/}


/*
     **********************************
	* DOS CONVERSION FUNCTIOS        *
	**********************************
*/
void outport(int port, int value)
{
  outb (value, port); /*  The port and value arguments are in the opposite order */
}

short inport(int port)
{
  return inb(port);
}


/*
	******************************
	* RADIX CONVERSION FUNCTIONS *
	******************************
*/

/*
	Conversion of decimal number (0...255) to 8 bit binary string.
	--------------------------------------------------------------
*/
void DecToBin(int DecNumber, char *Bin)
{
	int i;
	char *one = "1", *zero = "0";

	DecNumber <<= 8;				/*Shift dec. number 8 bits left*/
	if (DecNumber < 0)
		strcpy(Bin, one);
	else
		strcpy(Bin, zero);
	DecNumber <<= 1;
	for (i=1; i<8; i++) {
		if (DecNumber < 0)     /*Write highest bit into bin string*/
			strcat(Bin, one);
		else
			strcat(Bin, zero);
		DecNumber <<= 1;	/*Shift decimal number 1 bit left*/
	};
};


/*
	Conversion of binary string (max 8 bit) into decimal number.
	------------------------------------------------------------
*/
int BinToDec(char *BinNumber)
{
	int i, Weight, dec;

	Weight = 1;	/*Set weight factor at lowest bit*/
	dec = 0;							/*Reset decimal number*/
	for (i=strlen(BinNumber)-1;i>=0;i--) {			/*Convert all bits from bin. string*/
		dec = dec + (BinNumber[i] == '1') * Weight;	/*If bit=1 then add weigth factor*/
		Weight = Weight << 1;				/*Multiply weight factor by 2*/
	};
	return dec;	/*return result*/
};


/*
	Conversion of decimal number (0...255) to hexadecimal string (2 digits).
	------------------------------------------------------------------------
*/
void DecToHex(int DecNumber, char *Hex)
{
	strcpy(Hex,"0x");						/*Set length hex string at 2*/
	/*itoa(DecNumber / 16, Hex+2, 16);*/			/*MSD = dec. number / 16*/
	/*itoa(DecNumber % 16, Hex+3, 16);*/			/*LSD = remainder dec number / 16*/
};


/*
	Conversion of hexadecimal string (2 digits) to decimal number.
	--------------------------------------------------------------
*/
int HexToDec(char *HexNumber)
{
	int Strlength, dech = 0, decl;

	Strlength = strlen(HexNumber);			/*Calculate length of hex string*/
	switch(HexNumber[Strlength-1]) {
		case 'a': case 'A': decl = 10; break;
		case 'b': case 'B': decl = 11; break;
		case 'c': case 'C': decl = 12; break;
		case 'd': case 'D': decl = 13; break;
		case 'e': case 'E': decl = 14; break;
		case 'f': case 'F': decl = 15; break;
		default: decl = HexNumber[Strlength-1]-'0';
	};
	if (Strlength == 2) { 			/*If second digit exist convert it*/
		switch(HexNumber[0]) {
			case 'a': case 'A': dech = 10; break;
			case 'b': case 'B': dech = 11; break;
			case 'c': case 'C': dech = 12; break;
			case 'd': case 'D': dech = 13; break;
			case 'e': case 'E': dech = 14; break;
			case 'f': case 'F': dech = 15; break;
			default: dech = HexNumber[0]-'0';
		};
	};
	return ((dech << 4) + decl);				/*Shift MSD 4 bit left & add to LSD*/
};
/*
	*****************************************
	* UPDATE IOdata & IO ARRAY PROCEDURES *
	*****************************************

	Update the 'IOdata' & 'IO' array with data for selected chip.
	-------------------------------------------------------------
*/

void UpdateIOdataArray(int Chip_no, int Data)
{
	int Start_Channel, Channel;

	/*Update 'IOdata' array*/
	IOdata[Chip_no] = (IOdata[Chip_no] & IOconfig[Chip_no]);
	IOdata[Chip_no] = IOdata[Chip_no] | (Data & (~IOconfig[Chip_no]));
	/*Update 'IO' array*/
	Start_Channel = Chip_no * 8 + 1;	/*Calculate start channel*/
	for (Channel = 0; Channel <= 7; Channel++)	/*Test status 8 ch. of the IOchip*/
		IO[Start_Channel+Channel] = ((IOdata[Chip_no] & (0x01 << Channel)) != 0);
};

/*
	Clear the 'IOdata' & 'IO' array from the selected chip.
	-------------------------------------------------------
*/
void ClearIOdataArray(int Chip_no)
{
	int Start_Channel, Channel;

	/*Update 'IOdata' array*/
	IOdata[Chip_no] = IOdata[Chip_no] & IOconfig[Chip_no];
	/*Update 'IO' array*/
	Start_Channel = Chip_no * 8 + 1;	/*Calculate start channel*/
	for (Channel = 0; Channel <= 7; Channel++)	/*Test status 8 ch. of the IOchip*/
		IO[Start_Channel+Channel] = ((IOdata[Chip_no] & (0x01 << Channel)) != 0);
};

/*
	Setthe 'IOdata' & 'IO' array from the selected chip.
	----------------------------------------------------
*/
void SetIOdataArray(int Chip_no)
{
	int Start_Channel, Channel;

	/*Update 'IOdata' array*/
	IOdata[Chip_no] = IOdata[Chip_no] | ( ~IOconfig[Chip_no]);
	/*Update 'IO' array*/
	Start_Channel = Chip_no * 8 + 1;		/*Calculate start channel*/
	for (Channel = 0; Channel <= 7; Channel++)	/*Test status 8 ch. of the IOchip*/
		IO[Start_Channel+Channel] = ((IOdata[Chip_no] & (0x01 << Channel)) != 0);
};

/*
	Set the 'IOdata' & 'IO' array from the selected channel.
	--------------------------------------------------------
*/
void SetIOchArray(int Channel_no)
{
	int Chip_no, Channel, Data;

	Chip_no = (Channel_no - 1) / 8;		/*Calculate chip no*/
	Channel = (Channel_no - 1) % 8;		/*Calculate channel of IOchip*/
	Data = IOdata[Chip_no] | (0x01 << Channel);	/*Data for correct IOchannel to set*/
	UpdateIOdataArray(Chip_no, Data);		/*Update IOdata & IO array*/
};

/*
	Clear the 'IOdata' & 'IO' array from the selected channel.
	----------------------------------------------------------
*/
void ClearIOchArray(int Channel_no)
{
	int Chip_no, Channel, Data;

	Chip_no = (Channel_no - 1) / 8;		/*Calculate chip no*/
	Channel = (Channel_no - 1) % 8;		/*Calculate channel of IOchip*/
	Data = IOdata[Chip_no] & (~(0x01 << Channel));	/*Data for IOchannel to clear*/
	UpdateIOdataArray(Chip_no, Data);		/*Update IOdata & IO array*/
};
/*
	*********************
	* OUTPUT PROCEDURES *
	*********************

	Output data to selected IO-port and update the 'IOdata' & 'IO' array.
	---------------------------------------------------------------------
*/
void IOoutput(int Chip_no, int Data)
{
	short Start_Channel, Channel, SerData, cport, i, j;

	Data = (~Data) | IOconfig[Chip_no];	/*Mask input channels*/

	/*Generate start condition*/
	outport(ControlPort, 0x06);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x0E);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/


	/*Serial output of Chipcode*/
	SerData = (IOchipCode[Chip_no] << 8);     	/*Shift data 8 bits left*/
	for (j = 1; j <= 8; j++) {
		cport = (SerData < 0) ? 0x0C: 0x0E;
		outport(ControlPort, cport);
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		cport = cport & 0x07;
		outport(ControlPort, cport);
		SerData <<= 1;
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		cport = cport | 0x08;
		outport(ControlPort, cport);
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
	};

	/*Clock pulse for acknowledgement*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x04);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/

	/*Serial output of data*/
	SerData = (Data << 8);
	for (j = 1; j <= 8; j++) {
		cport = (SerData < 0) ? 0x0C: 0x0E;
		outport(ControlPort, cport);
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		cport = cport & 0x07;
		outport(ControlPort, cport);
		SerData <<= 1;
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		cport = cport | 0x08;
		outport(ControlPort, cport);
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
	};

	/*Clock pulse for acknowledgement*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x04);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/

	/*Generate stop condition*/
	outport(ControlPort, 0x0E);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x06);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x04);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/

	/*Update 'IOdata' array*/
	IOdata[Chip_no] = (IOdata[Chip_no] & IOconfig[Chip_no]) | (~Data);

	/*Update 'IO' array*/
	Start_Channel = Chip_no * 8 + 1;		/*Calculate start channel*/
	for (Channel = 0; Channel <= 7; Channel++)	/*Test status 8 ch. of the IOchip*/
		IO[Start_Channel+Channel] = ((IOdata[Chip_no] & (0x01 << Channel)) != 0);
};

/*
	Update all IO channels with value stored into the 'IOdata' array.
	-----------------------------------------------------------------
*/
void UpdateAllIO()
{
	int Chip_no;

	for (Chip_no = 0; Chip_no <= MaxIOchip; Chip_no++)
		IOoutput(Chip_no, IOdata[Chip_no]);	/*Update all IO-ports*/
};

/*
	Clear all IO channels and update the 'IOdata' & 'IO' array.
	-----------------------------------------------------------
*/
void ClearAllIO()
{
	int Chip_no;

	for (Chip_no = 0; Chip_no <= MaxIOchip; Chip_no++)
		IOoutput(Chip_no, 0);	/*Clear all IO-port*/
};

/*
	Set all IO channels and update the 'IOdata' & 'IO' array.
	---------------------------------------------------------
*/
void SetAllIO()
{
	int Chip_no;

	for (Chip_no = 0; Chip_no <= MaxIOchip; Chip_no++)
		IOoutput(Chip_no, 0x0FF);	/*Set all IO-port*/
};

/*
	Update IO chip with value stored into the 'IOdata' array.
	---------------------------------------------------------
*/
void UpdateIOchip(int Chip_no)
{
	IOoutput(Chip_no, IOdata[Chip_no]);	/*Update IO port*/
};

/*
	Clear IO chip and update the 'IOdata' & 'IO' array.
	---------------------------------------------------
*/
void ClearIOchip(int Chip_no)
{
	IOoutput(Chip_no, 0);	/*Clear IO-port*/
};

/*
	Set IO chip and update the 'IOdata' & 'IO' array.
	-------------------------------------------------
*/
void SetIOchip(int Chip_no)
{
	IOoutput(Chip_no, 0x0FF);	/*Set IO-port*/
};

/*
	Set one IO channel and update the 'IOdata' & 'IO' array.
	--------------------------------------------------------
*/
void SetIOchannel(int Channel_no)
{
	int Chip_no, Channel, Data;

	Chip_no = (Channel_no - 1) / 8;		/*Calculate chip no*/
	Channel = (Channel_no - 1) % 8;		/*Calculate channel of IOchip*/
	Data = IOdata[Chip_no] | (0x01 << Channel);	/*Data for correct IOchannel to set*/
	IOoutput(Chip_no, Data);			/*Set IOchannel*/
};

/*
	Clear one IO channel and update the 'IOdata' & 'IO' array.
	----------------------------------------------------------
*/
void ClearIOchannel(int Channel_no)
{
	int Chip_no, Channel, Data;

	Chip_no = (Channel_no - 1) / 8;		/*Calculate chip no*/
	Channel = (Channel_no - 1) % 8;		/*Calculate channel of IOchip*/
	Data = IOdata[Chip_no] & (~(0x01 << Channel));	/*Data for IOchannel to clear*/
	IOoutput(Chip_no, Data);			/*Clear IOchannel*/
};

/*
	*********************
	* INPUT PROCEDURES  *
	*********************

	Read IO chip and update the 'IOdata' & 'IO' array.
	--------------------------------------------------
*/
void ReadIOchip(int Chip_no)
{
	short Start_Channel, Channel, Data, cport, i, j;

	/*Generate start condition*/
	outport(ControlPort, 0x06);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x0E);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/

	/*Serial output of Chipcode for readmode*/
	Data = ((IOchipCode[Chip_no] | 0x01) << 8);  /*Shift data 8 bits left*/
	for (j = 1; j <= 8; j++) {
		cport = (Data < 0) ? 0x0C: 0x0E;
		outport(ControlPort, cport);
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		cport = cport & 0x07;
		outport(ControlPort, cport);
		Data <<= 1;
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		cport = cport | 0x08;
		outport(ControlPort, cport);
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
	};

	/*Clock pulse for acknowledgement*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x04);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/

	/*Serial input of ChipData*/
	Data = 0;				/*Clear Data*/
	for (j = 1; j <= 8; j++) {
		Data <<= 1;
		outport(ControlPort, 0x04);
		for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
		if (inport(StatusPort) & 0x10) Data = Data | 0x0001;
		outport(ControlPort, 0x0C);
		for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	};
	IOdata[Chip_no] = ~Data;

	/*Clock pulse for acknowledgement*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x04);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/

	/*Generate stop condition*/
	outport(ControlPort, 0x0E);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x06);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x04);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/

	/*Update 'IO' array*/
	Start_Channel = Chip_no * 8 + 1;	/*Calculate start channel*/
	for (Channel = 0; Channel <= 7; Channel++)	/*Test status 8 ch. of the IOchip*/
		IO[Start_Channel+Channel] = ((IOdata[Chip_no] & (0x01 << Channel)) != 0);
};

/*
	Read all IO channels and update the 'IOdata' & 'IO' array.
	----------------------------------------------------------
*/
void ReadAllIO()
{
	int Chip_no;

	for (Chip_no = 0; Chip_no <= MaxIOchip; Chip_no++)
		ReadIOchip(Chip_no);	/*Read all IO ports*/
};

/*
	Read one IO channels and update the 'IOdata' & 'IO' array.
	----------------------------------------------------------
*/
void ReadIOchannel(int Channel_no)
{
	int Chip_no;
	
	Chip_no = (Channel_no - 1) / 8;	/*Calculate chip no*/
	ReadIOchip(Chip_no);		/*Read IO port*/
};

/*	*******************************
	* IO CONFIGURATION PROCEDURES *
	*******************************

	Config all IO-ports as inputs (1 = Input mode / 0 = Output Mode).
	-----------------------------------------------------------------
*/
void ConfigAllIOasInput()
{
	int Chip_no;

	for (Chip_no = 0; Chip_no <= MaxIOchip; Chip_no++) {
		IOconfig[Chip_no] = 0;		/*Config all IO-ports as outputs*/
		ClearIOchip(Chip_no);		/*Clear all IO-ports*/
		IOconfig[Chip_no] = 0xFF;	/*Config all IO-ports as inputs*/
		ReadIOchip(Chip_no);		/*Update 'IOdata' & 'IO' array*/
	}
};

/*
	Config all IO-ports as outputs (1 = Input mode / 0 = Output Mode).
	-----------------------------------------------------------------
*/
void ConfigAllIOasOutput()
{
	int Chip_no;

	for (Chip_no = 0; Chip_no <= MaxIOchip; Chip_no++)
		IOconfig[Chip_no] = 0x00;	/*Config all IO-ports as outputs*/
	ClearAllIO();				/*Clear all IO-ports*/
};

/*
	Config one IO-port as input (1 = Input mode / 0 = Output mode).
	---------------------------------------------------------------
*/
void ConfigIOchipAsInput(int Chip_no)
{
	IOconfig[Chip_no] = 0;		/*Config IO-port as output*/
	ClearIOchip(Chip_no);		/*Clear  IO-port*/
	IOconfig[Chip_no] = 0xFF;	/*Config IO-port as inputs*/
	ReadIOchip(Chip_no);		/*Update 'IOdata' & 'IO' array*/
};

/*
	Config one IO-port as output (1 = Input mode / 0 = Output mode).
	----------------------------------------------------------------
*/
void ConfigIOchipAsOutput(int Chip_no)
{
	IOconfig[Chip_no] = 0x00;	/*IO-port as outputs*/
	ClearIOchip(Chip_no);		/*Clear IO-port*/
};

/*
	Config one IO-channel as input (1 = Input mode / 0 = Output mode).
	------------------------------------------------------------------
*/
void ConfigIOchannelAsInput(int Channel_no)
{
	int Chip_no, Channel;

	Chip_no = (Channel_no - 1) / 8;		/*Calculate chip no*/
	Channel = (Channel_no - 1) % 8;		/*Calculate channel from IOchip*/
	IOconfig[Chip_no] = IOconfig[Chip_no] & (~(0x01 << Channel)); /*Set IOchannel as output*/
	ClearIOchannel(Channel_no);			/*clear IO-channel*/
	IOconfig[Chip_no] = IOconfig[Chip_no] | (0x01 << Channel); /*Set IOchannel as input*/
	ReadIOchannel(Channel_no);			/*Update 'IOdata' & 'IO' array*/
};

/*
	Config one IO-channel as output (1 = Input mode / 0 = Output mode).
	-------------------------------------------------------------------
*/
void ConfigIOchannelAsOutput(int Channel_no)
{
	int Chip_no, Channel;

	Chip_no = (Channel_no - 1) / 8;		/*Calculate chip no*/
	Channel = (Channel_no - 1) % 8;		/*Calculate channel from IOchip*/
	IOconfig[Chip_no] = IOconfig[Chip_no] & (~(0x01 << Channel)); /*Set IOchannel as output*/
	ClearIOchannel(Channel_no);			/*clear IO-channel*/
};

/*
	**********************************
	* 6 BIT DAC CONVERTER PROCEDURES *
	**********************************

	Set selected DAC channel by given value. The 'DAC' array will be updated.
	-------------------------------------------------------------------------
*/
void OutputDACchannel(int Channel_no, int Data)
{
	short SerData, cport, i, j;

	Data = (Data > 63) ? 63: Data;	/*If overflow  set at maximum (63)*/
	DAC[Channel_no] = Data;	/*Update 'DAC' array*/

	/*Generate start condition*/
	outport(ControlPort, 0x06);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x0E);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/

	/*Serial output of Chipcode*/
	SerData = ((DACchipCode[(Channel_no - 1) / 8]) << 8);	/*Shift data 8 bits left*/
	for (j = 1; j <= 8; j++) {
		cport = (SerData < 0) ? 0x0C: 0x0E;
		outport(ControlPort, cport);
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		cport = cport & 0x07;
		outport(ControlPort, cport);
		SerData <<= 1;
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		cport = cport | 0x08;
		outport(ControlPort, cport);
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
	};

	/*Clock pulse for acknowledgement*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x04);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/

	/*Serial output of selected channel*/
	SerData = ((0xF0 | ((Channel_no - 1) % 8)) << 8);	/*Shift data 8 bits left*/
	for (j = 1; j <= 8; j++) {
		cport = (SerData < 0) ? 0x0C: 0x0E;
		outport(ControlPort, cport);
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		cport = cport & 0x07;
		outport(ControlPort, cport);
		SerData <<= 1;
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		cport = cport | 0x08;
		outport(ControlPort, cport);
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
	};

	/*Clock pulse for acknowledgement*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x04);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/

	/*Serial output of Data*/
	SerData = (Data << 8);
	for (j = 1; j <= 8; j++) {
		cport = (SerData < 0) ? 0x0C: 0x0E;
		outport(ControlPort, cport);
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		cport = cport & 0x07;
		outport(ControlPort, cport);
		SerData <<= 1;
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		cport = cport | 0x08;
		outport(ControlPort, cport);
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
	};

	/*Clock pulse for acknowledgement*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x04);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/

	/*Generate stop condition*/
	outport(ControlPort, 0x0E);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x06);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x04);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
};

/*
	Update the eight DAC channels from the selected chip by the 'DAC' array.
	------------------------------------------------------------------------
*/
void UpdateDACchip(int Chip_no)
{
	short SerData, i, j, Channel, k, cport;

	/*Generate start condition*/
	outport(ControlPort, 0x06);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x0E);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/

	/*Serial output of Chipcode*/
	SerData = (DACchipCode[Chip_no] << 8);    /*Shift data 8 bits left*/
	for (j = 1; j <= 8; j++) {
		cport = (SerData < 0) ? 0x0C: 0x0E;
		outport(ControlPort, cport);
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		cport = cport & 0x07;
		outport(ControlPort, cport);
		SerData <<= 1;
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		cport = cport | 0x08;
		outport(ControlPort, cport);
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
	};

	/*Clock pulse for acknowledgement*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x04);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/

	/*Serial output of channel 0 from selected chip*/
	SerData = (0x00);
	for (j = 1; j <= 8; j++) {
		cport = (SerData < 0) ? 0x0C: 0x0E;
		outport(ControlPort, cport);
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		cport = cport & 0x07;
		outport(ControlPort, cport);
		SerData <<= 1;
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		cport = cport | 0x08;
		outport(ControlPort, cport);
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
	};

	/*Clock pulse for acknowledgement*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x04);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/

	/*Update the 8 channels of the selected chip*/
	Channel = Chip_no * 8;				/*Calculate DAC channel*/
	for (k = 1; k <= 8; k++) {
		/*Serial output of DAC Data*/
		DAC[Channel + k] = (DAC[Channel+k] > 63) ? 63 : DAC[Channel+k];	/*Overflow*/
		/*Shift data 8 bits left*/
		SerData = (DAC[Channel + k] << 8);
		for (j = 1; j <= 8; j++) {
			cport = (SerData < 0) ? 0x0C: 0x0E;
			outport(ControlPort, cport);
			for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
			cport = cport & 0x07;
			outport(ControlPort, cport);
			SerData <<= 1;
			for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
			cport = cport | 0x08;
			outport(ControlPort, cport);
			for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		};

		/*Clock pulse for acknowledgement*/
		outport(ControlPort, 0x0C);
		for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
		outport(ControlPort, 0x04);
		for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
		outport(ControlPort, 0x0C);
		for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	};

	/*Generate stop condition*/
	outport(ControlPort, 0x0E);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x06);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x04);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
};

/*
	Update DAC channels by the 'DA' array.
	--------------------------------------
*/
void UpdateDACchannel(int Channel_no)
{
	OutputDACchannel(Channel_no, DAC[Channel_no]); /*Update DAC channel*/
};

/*
	Set selected DAC channel at minimum (0). The 'DAC' array will be updated.
	-------------------------------------------------------------------------
*/
void ClearDACchannel(int Channel_no)
{
	OutputDACchannel(Channel_no, 0); /*Set selected DAC channel at 0*/
};

/*
	Set selected DAC channel at maximum (63). The 'DAC' array will be updated.
	--------------------------------------------------------------------------
*/
void SetDACchannel(int Channel_no)
{
	OutputDACchannel(Channel_no,63);  /*Set selected DAC channel at 63*/
};

/*
	Set the eight DAC channels from the selected chip at minimum (0) and update the 'DAC' array.
	--------------------------------------------------------------------------------------------
*/
void ClearDACchip(int Chip_no)
{
	int i, Channel;

	Channel = Chip_no * 8;
	for (i = 1; i <= 8; i++)
		DAC[Channel + i] = 0;	/*Clear 8 channels of selected chip*/
	UpdateDACchip(Chip_no);		/*Update DAC chip*/
};

/*
	Set the eight DAC channels from the selected chip at maximum (63) and update the 'DAC' array.
	---------------------------------------------------------------------------------------------
*/
void SetDACchip(int Chip_no)
{
	int i, Channel;

	Channel = Chip_no * 8;
	for (i = 1; i <= 8; i++)
		DAC[Channel + i] = 63;	/*Set 8 channels selected chip max*/
	UpdateDACchip(Chip_no);		/*Update DAC chip*/
};

/*
	Update all DAC channels by the 'DAC' array.
	-------------------------------------------
*/
void UpdateAllDAC()
{
	int Chip_no;

	for (Chip_no = 0; Chip_no <= MaxIOcard; Chip_no++)
		UpdateDACchip(Chip_no);	/*Update all DAC channels*/
};

/*
	Set all DAC channels at minimum (0) and update the 'DAC' array.
	---------------------------------------------------------------
*/
void ClearAllDAC()
{
	int Channel_no;

	for (Channel_no = 1; Channel_no <= MaxDACchannel; Channel_no++)
		DAC[Channel_no] = 0x00;	/*Clear the 'DAC' array*/
	UpdateAllDAC();				/*Update all DAC channels*/
};

/*
	Set all DAC channels at maximum (63) and update the 'DAC' array.
	----------------------------------------------------------------
*/
void SetAllDAC()
{
	int Channel_no;

	for (Channel_no = 1; Channel_no <= MaxDACchannel; Channel_no++)
		DAC[Channel_no] = 63;	/*Set 'DAC' array at max.(63)*/
	UpdateAllDAC();			/*Update all DAC channels*/
};

/*
	*********************************
	* 8 BIT DA CONVERTER PROCEDURES *
	*********************************

	Set selected DA channel by given value. The 'DA' array will be updated.
	-----------------------------------------------------------------------
*/
void OutputDAchannel(int Channel_no, int Data)
{
	short SerData, cport, i, j;

	DA[Channel_no] = Data;	/*Store DA data into DA array*/

	/*Generate start condition*/
	outport(ControlPort, 0x06);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x0E);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/

	/*Serial output of Chipcode*/
	SerData = ((ADDAchipCode[Channel_no-1] << 8));	/*Shift data 8 bits left*/
	for (j = 1; j <= 8; j++) {
		cport = (SerData < 0) ? 0x0C: 0x0E;
		outport(ControlPort, cport);
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		cport = cport & 0x07;
		outport(ControlPort, cport);
		SerData <<= 1;
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		cport = cport | 0x08;
		outport(ControlPort, cport);
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
	};

	/*Clock pulse for acknowledgement*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x04);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/

	/*Serial output of ControlByte*/
	SerData = (0x40 << 8);
	for (j = 1; j <= 8; j++) {
		cport = (SerData < 0) ? 0x0C: 0x0E;
		outport(ControlPort, cport);
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		cport = cport & 0x07;
		outport(ControlPort, cport);
		SerData <<= 1;
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		cport = cport | 0x08;
		outport(ControlPort, cport);
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
	};

	/*Clock pulse for acknowledgement*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x04);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/

	/*Serial output of DA Data*/
	SerData = (Data << 8);
	for (j = 1; j <= 8; j++) {
		cport = (SerData < 0) ? 0x0C: 0x0E;
		outport(ControlPort, cport);
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		cport = cport & 0x07;
		outport(ControlPort, cport);
		SerData <<= 1;
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		cport = cport | 0x08;
		outport(ControlPort, cport);
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
	};

	/*Clock pulse for acknowledgement*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x04);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/

	/*Generate stop condition*/
	outport(ControlPort, 0x0E);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x06);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x04);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
};

/*
	Update DA channels by the 'DA' array .
	--------------------------------------
*/
void UpdateDAchannel(int Channel_no)
{
	OutputDAchannel(Channel_no, DA[Channel_no]);	/*Update DA channel */
};

/*
	Set selected DA channel at minimum (0). The 'DA' array will be updated.
	-----------------------------------------------------------------------
*/
void ClearDAchannel(int Channel_no)
{
	OutputDAchannel(Channel_no, 0);	/*Set selected DA channel at 0*/
};

/*
	Set selected DA channel at maximum (255). The 'DA' array will be updated.
	-------------------------------------------------------------------------
*/
void SetDAchannel(int Channel_no)
{
	OutputDAchannel(Channel_no, 255);	/*Set selected DA channel at 255*/
};

/*
	Update all DA channels by the 'DA' array .
	------------------------------------------
*/
void UpdateAllDA()
{
	int Channel_no;

	for (Channel_no = 1; Channel_no <= MaxDAchannel; Channel_no++)
		OutputDAchannel(Channel_no, DA[Channel_no]);	/*Update all DA channels*/
};

/*
	Set all DA channels at minimum (0). The 'DA' array will be cleared.
	-------------------------------------------------------------------
*/
void ClearAllDA()
{
	int channel_no;

	for (channel_no = 1; channel_no <= MaxDAchannel; channel_no++)
		OutputDAchannel(channel_no, 0);	/*Set all DA channels at 0*/
};

/*
	Set all DA channels at maximum (255). The 'DA' array will be updated.
	---------------------------------------------------------------------
*/
void SetAllDA()
{
	int Channel_no;

	for (Channel_no = 1; Channel_no <= MaxDAchannel; Channel_no++ )
		OutputDAchannel(Channel_no, 255);			/*Set all DA channels at 255*/
};

/*
	*********************************************
	* 8 bit AD CONVERTER FUNCTIONS & PROCEDURES *
	*********************************************

	Read one AD channel. The result will be stored into the 'AD' array.
	-------------------------------------------------------------------
*/
void ReadADchannel(int Channel_no)
{
	short ChipCode,i,j,cport,Data;

	ChipCode = ADDAchipCode[(Channel_no - 1) / 4];	/*Calculate chipcode*/

	/*Generate start condition*/
	outport(ControlPort, 0x06);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x0E);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/

	/*Serial output of Chipcode*/
	Data = (ChipCode << 8);				/*Shift data 8 bits left*/
	for (j = 1; j <= 8; j++) {
		cport = (Data < 0) ? 0x0C: 0x0E;
		outport(ControlPort, cport);
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		cport = cport & 0x07;
		outport(ControlPort, cport);
		Data <<= 1;
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		cport = cport | 0x08;
		outport(ControlPort, cport);
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
	};


	/*Clock pulse for acknowledgement*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x04);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/

	/*Serial output of selected channel*/
	Data = ((0x40 | ((Channel_no - 1) % 4)) << 8);	/*Shift data 8 bits left*/
	for (j = 1; j <= 8; j++) {
		cport = (Data < 0) ? 0x0C: 0x0E;
		outport(ControlPort, cport);
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		cport = cport & 0x07;
		outport(ControlPort, cport);
		Data <<= 1;
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		cport = cport | 0x08;
		outport(ControlPort, cport);
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
	};

	/*Clock pulse for acknowledgement*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x04);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/

	/*Generate stop condition*/
	outport(ControlPort, 0x0E);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x06);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x04);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/

	/*Generate start condition*/
	outport(ControlPort, 0x06);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x0E);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/

	/*Serial output of Chipcode for readmode*/
	Data = ((ChipCode | 0x01) << 8);		        /*Shift data 8 bits left*/
	for (j = 1; j <= 8; j++) {
		cport = (Data < 0) ? 0x0C: 0x0E;
		outport(ControlPort, cport);
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		cport = cport & 0x07;
		outport(ControlPort, cport);
		Data <<= 1;
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		cport = cport | 0x08;
		outport(ControlPort, cport);
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
	};

	/*Clock pulse for acknowledgement*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x04);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/

	/*Serial input of previous converted byte*/
	Data = 0;				/*Clear Data*/
	for (j = 1; j <= 8; j++) {
		Data <<= 1;
		outport(ControlPort, 0x04);
		for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
		if (inport(StatusPort) & 0x10) Data = Data | 0x0001;
		outport(ControlPort, 0x0C);
		for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	};

	/*Clock pulse for acknowledgement from master*/
	outport(ControlPort, 0x0E);
	for (i = 0; i <= I2CbusDelay; i++) {};     /*I2c-bus timing*/
	outport(ControlPort, 0x06);
	for (i = 0; i <= I2CbusDelay; i++) {};     /*I2c-bus timing*/
	outport(ControlPort, 0x0E);
	for (i = 0; i <= I2CbusDelay; i++) {};     /*I2c-bus timing*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};     /*I2c-bus timing*/

	/*Serial input of current converted byte*/
	Data = 0;				/*Clear Data*/
	for (j = 1; j <= 8; j++) {
		Data <<= 1;
		outport(ControlPort, 0x04);
		for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
		if (inport(StatusPort) & 0x10) Data = Data | 0x0001;
		outport(ControlPort, 0x0C);
		for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	};
	AD[Channel_no] = Data;			/*Store result*/

	/*Clock pulse for acknowledgement*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x04);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/

	/*Generate stop condition*/
	outport(ControlPort, 0x0E);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x06);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x04);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
};

/*
	Read four AD channels from one chip. The result will be stored into the 'AD' array.
	-----------------------------------------------------------------------------------
*/
void ReadADchip(int Chip_no)
{
	short Channel, k, j, i, Data, cport;

	Channel = Chip_no * 4 + 1;	/*Calculate first channel*/
																										/*Generate start condition*/
	/*Generate start condition*/
	outport(ControlPort, 0x06);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x0E);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/

	/*Serial output of Chipcode*/
	Data = (ADDAchipCode[Chip_no] << 8);	/*Shift data 8 bits left*/
	for (j = 1; j <= 8; j++) {
		cport = (Data < 0) ? 0x0C: 0x0E;
		outport(ControlPort, cport);
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		cport = cport & 0x07;
		outport(ControlPort, cport);
		Data <<= 1;
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		cport = cport | 0x08;
		outport(ControlPort, cport);
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
	};

	/*Clock pulse for acknowledgement*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x04);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/

	/*Serial output of controlbyte for autoincrement*/
	Data = (0x44 << 8);
	for (j = 1; j <= 8; j++) {
		cport = (Data < 0) ? 0x0C: 0x0E;
		outport(ControlPort, cport);
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		cport = cport & 0x07;
		outport(ControlPort, cport);
		Data <<= 1;
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		cport = cport | 0x08;
		outport(ControlPort, cport);
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
	};

	/*Clock pulse for acknowledgement*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x04);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/

	/*Generate stop condition*/
	outport(ControlPort, 0x0E);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x06);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x04);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/

	/*Generate start condition*/
	outport(ControlPort, 0x06);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x0E);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/

	/*Serial output of Chipcode for readmode*/
	Data = ((ADDAchipCode[Chip_no] | 0x01) << 8);	/*Shift data 8 bits left*/
	for (j = 1; j <= 8; j++) {
		cport = (Data < 0) ? 0x0C: 0x0E;
		outport(ControlPort, cport);
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		cport = cport & 0x07;
		outport(ControlPort, cport);
		Data <<= 1;
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
		cport = cport | 0x08;
		outport(ControlPort, cport);
		for (i = 0; i <= I2CbusDelay; i++) {};  /*I2c-bus timing*/
	};

	/*Clock pulse for acknowledgement*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x04);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/

	/*Serial input of previous converted byte*/
	Data = 0;				/*Clear Data*/
	for (j = 1; j <= 8; j++) {
		Data <<= 1;
		outport(ControlPort, 0x04);
		for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
		if (inport(StatusPort) & 0x10) Data = Data | 0x0001;
		outport(ControlPort, 0x0C);
		for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	};

	/*Read 4 AD channels*/
	for (k = 0; k <= 3; k++) {
		/*ACKnowledge from MASTER*/
		outport(ControlPort, 0x0E);
		for (i = 0; i <= I2CbusDelay; i++) {};     /*I2c-bus timing*/
		outport(ControlPort, 0x06);
		for (i = 0; i <= I2CbusDelay; i++) {};     /*I2c-bus timing*/
		outport(ControlPort, 0x0E);
		for (i = 0; i <= I2CbusDelay; i++) {};     /*I2c-bus timing*/
		outport(ControlPort, 0x0C);
		for (i = 0; i <= I2CbusDelay; i++) {};     /*I2c-bus timing*/
		/*Serial input of current converted byte*/
		Data = 0;				/*Clear Data*/
		for (j = 1; j <= 8; j++) {
			Data <<= 1;
			outport(ControlPort, 0x04);
			for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
			if (inport(StatusPort) & 0x10) Data = Data | 0x0001;
			outport(ControlPort, 0x0C);
			for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
		};
		AD[Channel + k] = Data;	/*Store data*/
	};

	/*Clock pulse for acknowledgement*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x04);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x0C);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/

	/*Generate stop condition*/
	outport(ControlPort, 0x0E);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x06);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
	outport(ControlPort, 0x04);
	for (i = 0; i <= I2CbusDelay; i++) {};    /*I2c-bus timing*/
};

/*
	Read all AD channels. The result will be stored into the 'AD' array.
	--------------------------------------------------------------------
*/
void ReadAllAD()
{
	int Chip_no;
					/*Scanning of all AD chips*/
	for (Chip_no = 0; Chip_no <= MaxIOcard; Chip_no++)
		ReadADchip(Chip_no);	/*Read 4 AD channels from chip*/
};

/*
	****************************
	* COMPLETE CARD PROCEDURES *
	****************************

	Read all IO's & AD's and update 'IOdata', 'IO' & 'AD' arrays.
  -------------------------------------------------------------
*/
void ReadAll()
{
	ReadAllIO();	/*Read all IO ports*/
	ReadAllAD();	/*Read all DA channels*/
};

/*
	Read all IO's & AD's from one card and update 'IOdata', 'IO' & 'AD' arrays.
  ---------------------------------------------------------------------------
*/
void ReadCard(int Card_no)
{
	int Chip_no;

	Chip_no = Card_no * 2;		/*Calculate IO chip no*/
	ReadIOchip(Chip_no);		/*Read first  8 IO ports*/
	ReadIOchip(Chip_no + 1);	/*Read second 8 IO ports*/
	ReadADchip(Card_no);		/*Read 4 DA channels*/
};

/*
	Update all IO's, DAC's & DA's and update 'IOdata', 'IO', 'DAC' & 'DA' arrays.
	-----------------------------------------------------------------------------
*/
void UpdateAll()
{
	UpdateAllIO();	/*Update all IO ports*/
	UpdateAllDAC();	/*Update all DAC channels*/
	UpdateAllDA();	/*Update all DA channels*/
};

/*
	Update all IO DAC & DA's from one card and update 'IOdata' 'IO' 'DAC' 'DA' arrays.
  ----------------------------------------------------------------------------------
*/
void UpdateCard(int Card_no)
{
	int Chip_no;

	Chip_no = Card_no * 2;		/*Calculate chip no*/
	UpdateIOchip(Chip_no);		/*Update first  8 IO ports*/
	UpdateIOchip(Chip_no + 1);	/*Update second 8 IO ports*/
	UpdateDACchip(Card_no);		/*Update 8 DAC channels*/
	UpdateDAchannel(Card_no + 1);	/*Update DA channel*/
};

/*
	**********************
	* I2C-BUS CONDITIONS *
	**********************

	SDA IN  =  SELECT    (pin 13) = bit 4 of STATUSPORT
	SDA OUT = -AUTOFEED  (pin 14) = bit 1 of CONTROLPORT
	SCL OUT = -SELECT IN (pin 17) = bit 3 of CONTROLPORT bits 0,4...7 = 0 / bit 2 = 1  of CONTROLPORT

	Set I2C-bus at non actif.
	-------------------------
*/
void I2CBusNotBusy()
{
	outport(ControlPort, 0x04);	/*SDA & SCL HIGH*/
};

/*
	Select I2C communication printerport.
	-------------------------------------
*/

void SelectI2CprinterPort(int Printer_no)
{
	switch (Printer_no) {
		case 0: /*lpt on monochroom display adapter*/
			StatusPort  = 0x03BD;
			ControlPort = 0x03BE;
			break;
		case 1:	/*lpt1 on mainboard*/
			StatusPort  = 0x0379;
			ControlPort = 0x037A;
			break;
		case 2:	/*lpt2 on mainboard*/
			StatusPort  = 0x0279;
			ControlPort = 0x027A;
			break;
	};
};

/*
	***********************
	* INITIALIZATION PART *
	***********************
*/
void initialize(void)
{
	SelectI2CprinterPort(1);	/*Select LPT1 on mainboard*/
	I2CbusDelay = 1;		/*I2C-bus speed control factor*/
	I2CBusNotBusy();

};


